"""Text"""

from . import client
import warnings
from . import utils

## Text renderer

counter = 0
class Text:
  def __init__(self, text = "", color = "#FFFFFF", font_size = 12, position = [0,0]):
    self.create()

    self.set_text(text)

    self.set_font_size(font_size)

    color = utils.sanitize_color(color)
    self.color = color
    client.sio.emit('SetTextColors',{self.id: color})

    position = utils.sanitize_list(position)
    self.position = position
    client.sio.emit('SetTextPositions',{self.id: position})


  def create(self):
    """Create a text object

    Parameters
    ----------
    none
        
    Examples
    --------
    >>> t1 = urchin.text.Text()
    """
    global counter
    counter +=1
    self.id = 't' + str(counter)
    client.sio.emit('CreateText',[self.id])
    self.in_unity = True
  
  def delete(self):
    """Delete a text object

    Parameters
    ----------
    references object being deleted
        
    Examples
    --------
    >>> t1.delete()
    """
    client.sio.emit('DeleteText',[self.id])
    self.in_unity = False

  def set_text(self, text_text):
    """Set the text in a set of text objects

    Parameters
    ----------
    text : string
      text to be displayed

    Examples
    --------
    >>> t1.set_text('test text')
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    text_text = utils.sanitize_string(text_text)
    self.text = text_text
    client.sio.emit('SetTextText',{self.id: text_text})

  def set_color(self,text_color):
    """Set the color of a set of text objects

    Parameters
    ----------
    text_colors : string hex color
        hex colors as strings
        
    Examples
    --------
    >>> t1.set_color('#FF0000')
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    
    text_color = utils.sanitize_color(text_color)
    self.color = text_color
    client.sio.emit('SetTextColors',{self.id: text_color})

  def set_font_size(self,text_size):
    """Set the font size of a set of text objects

    Parameters
    ----------
    text_sizes : int
        font sizes
        
    Examples
    --------
    >>> t1.set_size(12)
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    self.size = text_size
    client.sio.emit('SetTextSizes',{self.id: text_size})

  def set_position(self,position):
    """Set the positions of a set of text objects in UI canvas space
    Bottom left corner is [-1,-1], top right [1,1]

    Text is anchored at the top left corner of its text box.

    Parameters
    ----------
    text_pos : list of two floats
        canvas positions relative to the center
        
    Examples
    --------
    >>> t1.set_position([400, 300])
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    self.position = utils.sanitize_list(position)
    client.sio.emit('SetTextPositions',{self.id: self.position})


def create(n):
  """Create n text objects with default parameters

  Parameters
  ----------
  n : int
      number of text objects
  """
  text_list = []
  for i in range(n):
    text_list.append(Text())
  return text_list

def set_texts(text_list, str_list):
  """Set the string value of multiple text objects

  Parameters
  ----------
  text_list : list of Text
      Text objects
  str_list : _type_
      _description_
  """
  text_list = utils.sanitize_list(text_list)
  str_list = utils.sanitize_list(str_list, len(text_list))

  text_strs = {}
  for i, text in enumerate(text_list):
    text_strs[text.id] = str_list[i]
  
  client.sio.emit('SetTextText',text_strs)

def set_positions(text_list, pos_list):
  """Set the positions of multiple text objects

  Positions are [0,1] relative to the edges of the screen

  Parameters
  ----------
  text_list : list of Text
      Text objects
  pos_list : list of float
      [0,0] top left [1,1] bottom right
  """
  text_list = utils.sanitize_list(text_list)
  pos_list = utils.sanitize_list(pos_list, len(text_list))

  text_poss = {}
  for i, text in enumerate(text_list):
    text_poss[text.id] = pos_list[i]
  
  client.sio.emit('SetTextPositions',text_poss)

def set_font_sizes(text_list, font_size_list):
  """_summary_

  Parameters
  ----------
  text_list : list of Text
      Text objects
  font_size_list : _type_
      _description_
  """
  text_list = utils.sanitize_list(text_list)
  font_size_list = utils.sanitize_list(font_size_list, len(text_list))

  text_font_sizes = {}
  for i, text in enumerate(text_list):
    text_font_sizes[text.id] = font_size_list[i]
  
  client.sio.emit('SetTextSizes',text_font_sizes)

def set_colors(text_list, color_list):
  """_summary_

  Parameters
  ----------
  text_list : list of Text
      Text objects
  color_list : _type_
      _description_
  """
  text_list = utils.sanitize_list(text_list)
  color_list = utils.sanitize_list(color_list, len(text_list))

  text_colors = {}
  for i, text in enumerate(text_list):
    text_colors[text.id] = color_list[i]
  
  client.sio.emit('SetTextColors',text_colors)


"""Probers"""

from . import client
import warnings
from . import utils

##Probes Renderer
counter = 0
class Probe:
	def __init__(self, color = 'FFFFFF', position = [0,0,0], angle = [0,0,0], style = 'line', scale = [0.070, 3.840, 0.020]):
		self.create()

		color = utils.sanitize_color(color)
		self.color = color
		client.sio.emit('SetProbeColors', {self.id:color})

		position = utils.sanitize_vector3(position)
		self.position = position
		client.sio.emit('SetProbePos', {self.id:position})

		angle = utils.sanitize_vector3(angle)
		self.angle = angle
		client.sio.emit('SetProbeAngles', {self.id:angle})
		
		style = utils.sanitize_string(style)
		self.style = style
		#client.sio.emit('SetProbeStyle', {self.id:style})

		scale = utils.sanitize_vector3(scale)
		self.scale = scale
		client.sio.emit('SetProbeSize', {self.id:scale})

	def create(self):
		"""Create probe objects

		Parameters
		----------
		none

		Examples
		--------
		>>> p1 = urchin.probes.Probe()
		"""
		global counter
		counter +=1
		self.id = 'p' + str(counter)
		client.sio.emit('CreateProbes', [self.id])
		self.in_unity = True

	def delete(self):
		"""Delete probe objects

		Parameters
		----------
		references probe being deleted

		Examples
		--------
		>>> p1.delete()
		"""
		client.sio.emit('DeleteProbes', [self.id])
		self.in_unity = False

	def set_color(self,color):
		"""Set colors of probe objects

		Parameters
		----------
		color : string
			string is hex color

		Examples
		--------
		>>> p1.set_color('#FFFFFF')
		"""
		if self.in_unity == False:
			raise Exception("Object does not exist in Unity, call create method first.")
		
		color = utils.sanitize_color(color)
		self.color
		client.sio.emit('SetProbeColors', {self.id:color})

	def set_position(self, probe_positions):
		"""Set probe tip position in AP/ML/DV coordinates in um relative to the zero coordinate (front, top, left)

		Parameters
		----------
		probe_positions : vector3
			value is list of floats in AP/ML/DV in um

		Examples
		--------
		>>> p1.set_position([500,1500,2500])
		"""
		if self.in_unity == False:
			raise Exception("Object does not exist in Unity, call create method first.")
		
		self.position = utils.sanitize_vector3(probe_positions)
		client.sio.emit('SetProbePos', {self.id:[self.position[0]/1000, self.position[1]/1000, self.position[2]/1000]})

	def set_angle(self, probe_angles):
		"""Set probe azimuth/elevation/spin angles in degrees

		Azimuth 0 = has the probe facing the AP axis, positive values rotate clockwise
		Elevation 0 = probe is vertical, 90 = horizontal

		Parameters
		----------
		probe_angles : float list
			value is list of floats in az/elev/spin
			
		Examples
		--------
		>>> p1.set_angle([-90,0,0])
		"""
		if self.in_unity == False:
			raise Exception("Object does not exist in Unity, call create method first.")
		probe_angles = utils.sanitize_vector3(probe_angles)
		self.angle = probe_angles
		client.sio.emit('SetProbeAngles', {self.id:probe_angles})

	# def set_probe_style(self,probe_data):
	# 	"""Set probe rendering style

	# 	Style options are:
	# 		"line"
	# 		"probe-tip"
	# 		"probe-silicon"
	# 		"probe"

	# 	Inputs:
	# 	probe_data -- dictionary of probe names and string {'p1':'line'}
	# 	"""
	# 	if self.in_unity == False:
	# 		raise Exception("Object does not exist in Unity, call create method first.")
		
	# 	probe_data = utils.sanitize_string(probe_data)
	# 	self.style = probe_data
	# 	client.sio.emit('SetProbeStyle', {self.id:probe_data})

	def set_scale(self, probe_scale):
		"""Set probe scale in mm units, by default probes are scaled to 70 um wide x 20 um deep x 3840 um tall which is the size of a NP 1.0 probe.

		Parameters
		----------
		probe_scale: float list
			list of floats for width, height, depth
			
		Examples
		--------
		>>> p1.set_scale([0.070, 3.840, 0.020])
		"""
		if self.in_unity == False:
			raise Exception("Object does not exist in Unity, call create method first.")
		probe_scale = utils.sanitize_vector3(probe_scale)
		self.scale = probe_scale
		client.sio.emit('SetProbeSize', {self.id:probe_scale})

	
def create(num_objects):
	"""Create probe objects

	Parameters
	----------
	num_objects: int
		number of probe objects to be created

	Examples
	--------
	>>> probes = urchin.probes.create(2)
	"""
	probe_list = []
	for i in range(num_objects):
		probe_list.append(Probe())
	return probe_list

def delete(probes_list):
	"""Delete probe objects

	Parameters
	----------
	probes_list: list
		list of probe objects to be deleted

	Examples
	--------
	>>> probes.delete([p1,p2])
	"""
	probes_list = utils.sanitize_list(probes_list)
	probe_ids = [x.id for x in probes_list]
	client.sio.emit('DeleteProbes', probe_ids)

def set_colors(probes_list, colors_list):
	"""Set colors of probe objects

	Parameters
	----------
	probes_list: list of probe objects
		list of probe objects to be colored
	colors_list : list of string hex colors
		new hex colors for each probe

	Examples
	--------
	>>> urchin.probes.set_colors(probes,['#FFFFFF','#000000'])
	"""
	probes_list = utils.sanitize_list(probes_list)
	colors_list = utils.sanitize_list(colors_list)

	probe_colors = {}
	for i in range(len(probes_list)):
		probe = probes_list[i]
		if probe.in_unity:
			probe_colors[probe.id] = utils.sanitize_color(colors_list[i])
		else:
			warnings.warn(f"Object with id {probe.id} does not exist in Unity, Please create object {probe.id}.")
	client.sio.emit('SetProbeColors', probe_colors)

def set_positions(probes_list, positions_list):
	"""Set probe tip positions in AP/ML/DV coordinates in um relative to the zero point (front, left, top)

	Parameters
	----------
	probes_list : list of Probe
	positions_list : list of vector3
		tip coordinate in AP/ML/DV in um
      
	Examples
	--------
	>>> urchin.probes.set_positions(probes,[[1000,2000,1000],[2000,2000,2000]])
	"""
	probes_list = utils.sanitize_list(probes_list)
	positions_list = utils.sanitize_list(positions_list)

	probe_pos = {}
	for i in range(len(probes_list)):
		probe = probes_list[i]
		if probe.in_unity:
			pos = utils.sanitize_vector3(positions_list[i])
			probe_pos[probe.id] = [pos[0]/1000, pos[1]/1000, pos[2]/1000]
		else:
			warnings.warn(f"Object with id {probe.id} does not exist. Please create object {probe.id}.")

	client.sio.emit('SetProbePos', probe_pos)

def set_angles(probes_list, angles_list):
	"""Set probe azimuth/elevation/spin angles in degrees

	Azimuth 0 = has the probe facing the AP axis, positive values rotate clockwise
	Elevation 0 = probe is vertical, 90 = horizontal

	Parameters
	----------
	probes_list : list of probe objects
		list of probes being set
	probe_angles : list of list of three floats
		value is list of floats in az/elev/spin	
		
	Examples
	--------
	>>> urchin.probes.set_angles(probes,[[-90,0,0],[0,30,0]])
	"""
	probes_list = utils.sanitize_list(probes_list)
	angles_list = utils.sanitize_list(angles_list)

	probe_angle = {}
	for i in range(len(probes_list)):
		probe = probes_list[i]
		if probe.in_unity:
			probe_angle[probe.id] = utils.sanitize_vector3(angles_list[i])
		else:
			warnings.warn(f"Object with id {probe.id} does not exist. Please create object {probe.id}.")

	client.sio.emit('SetProbeAngles', probe_angle)

# def set_probe_styles(probes_list,styles_list):
# 	"""Set probe rendering style

# 	Style options are:
# 		"line"
# 		"probe-tip"
# 		"probe-silicon"
# 		"probe"

# 	Parameters
# 	----------
# 	probes_list : list of probe objects
# 		list of probes being set
# 	styles_list :list of strings
# 		list of strings for probe style
	
# 	Examples
# 	--------
# 	>>> urchin.probes.set_probe_styles(probes,['line','probe-tip'])
# 	"""
# 	probes_list = utils.sanitize_list(probes_list)
# 	styles_list = utils.sanitize_list(styles_list)

# 	probe_styles = {}
# 	for i in range(len(probes_list)):
# 		probe = probes_list[i]
# 		if probe.in_unity:
# 			probe_styles[probe.id] = utils.sanitize_string(styles_list[i])
# 		else:
# 			warnings.warn(f"Object with id {probe.id} does not exist in Unity, Please create object {probe.id}.")
# 	client.sio.emit('SetProbeStyle', probe_styles)

def set_scales(probes_list, scales_list):
	"""Set probe scale in mm units, by default probes are scaled to 70 um wide x 20 um deep x 3840 um tall which is the size of a NP 1.0 probe.

	Parameters
	----------
	probes_list: list of probe objects
		list of probe sizes being set
	scales_list: list of list of three floats
		list of floats for width, height, depth for each probe
		
	Examples
	--------
	>>> urchin.probes.set_scales(probes,[[0.070, 3.840, 0.020],[0.070, 3.840, 0.020]])
	"""
	probes_list = utils.sanitize_list(probes_list)
	scales_list = utils.sanitize_list(scales_list)

	probe_scale = {}
	for i in range(len(probes_list)):
		probe = probes_list[i]
		if probe.in_unity:
			probe_scale[probe.id] = utils.sanitize_vector3(scales_list[i])
		else:
			warnings.warn(f"Object with id {probe.id} does not exist. Please create object {probe.id}.")

	client.sio.emit('SetProbeSize', probe_scale)


"""Camera"""

from . import client
from . import utils

import PIL
from PIL import Image
import io
import json
import asyncio
		  
receive_totalBytes = {}
receive_bytes = {}
receive_camera = {}

PIL.Image.MAX_IMAGE_PIXELS = 22500000

# Handle receiving camera images back as screenshots
def on_camera_img_meta(data_str):
	"""Handler for receiving metadata about incoming images

	Parameters
	----------
	data_str : string
		JSON with two fields {"name":"", "totalBytes":""}
	"""
	global receive_totalBytes

	data = json.loads(data_str)

	name = data["name"]
	totalBytes = data["totalBytes"]

	receive_totalBytes[name] = totalBytes
	receive_bytes[name] = bytearray()

def on_camera_img(data_str):
	"""Handler for receiving data about incoming images

	Parameters
	----------
	data_str : string
		JSON with two fields {"name":"", "bytes":""}
	"""
	global receive_totalBytes, receive_bytes, receive_camera

	data = json.loads(data_str)

	name = data["name"]
	byte_data = bytes(data["data"])

	receive_bytes[name] = receive_bytes[name] + byte_data
	
	if len(receive_bytes[name]) == receive_totalBytes[name]:
		print(f'(Camera receive) Camera {name} received an image')
		receive_camera[name].image_received = True

## Camera renderer
counter = 0

class Camera:
	def __init__(self, main = False):		
		if main:
			self.id = 'CameraMain'
		else:
			global counter
			counter += 1
			self.id = f'Camera{counter}'
			client.sio.emit('CreateCamera', [self.id])
		self.in_unity = True
		self.image_received_event = asyncio.Event()
		self.loop = asyncio.get_event_loop()

		self.background_color = '#ffffff'

	def create(self):
		"""Creates camera
		
		Parameters
		---------- 
		none

		Examples
		>>>c1 = urchin.camera.Camera()
		"""

	def delete(self):
		"""Deletes camera
		
		Parameters
		---------- 
		references object being deleted

		Examples
		>>>c1.delete()
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		client.sio.emit('DeleteCamera', [self.id])
		self.in_unity = False

	def set_target_coordinate(self,camera_target_coordinate):
		"""Set the camera target coordinate in CCF space in um relative to CCF (0,0,0), without moving the camera. Coordinates can be negative or larger than the CCF space (11400,13200,8000)

		Parameters
		----------
		camera_target_coordinate : float list
		  list of coordinates in ap, ml, dv in um

		Examples
		--------
		>>>c1.set_target_coordinate([500,1500,1000])
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		camera_target_coordinate = utils.sanitize_vector3(camera_target_coordinate)
		self.target = camera_target_coordinate
		client.sio.emit('SetCameraTarget', {self.id: camera_target_coordinate})

	# temporarily removed
	# def set_position(self, position, preserve_target = True):
	# 	"""Set the camera position in CCF space in um relative to CCF (0,0,0), coordinates can be outside CCF space. 

	# 	Parameters
	# 	----------
	# 	position : float list
	# 		list of coordinates in ml/ap/dv in um
	# 	preserve_target : bool, optional
	# 		when True keeps the camera aimed at the current target, when False preserves the camera rotation, by default True
		
	# 	Examples
	# 	--------
	# 	>>> c1.set_position([500,1500,1000])
	# 	"""
	# 	if self.in_unity == False:
	# 		raise Exception("Camera is not created. Please create camera before calling method.")
		
	# 	position = utils.sanitize_vector3(position)
	# 	self.position = position
	# 	packet = position.copy()
	# 	packet.append(preserve_target)
	# 	client.sio.emit('SetCameraPosition', {self.id: packet})

	def set_rotation(self, rotation):
		"""Set the camera rotation (pitch, yaw, roll). The camera is locked to a target, so this rotation rotates around the target.

		Rotations are applied in order: roll, yaw, pitch. This can cause gimbal lock.

		Parameters
		----------
		rotation : float list OR string
			list of euler angles to set the camera rotation in (pitch, yaw, roll)
			OR
			string: "axial", "coronal", "sagittal", or "angled"

		Examples
		--------
		>>> c1.set_rotation([0,0,0])
		>>> c1.set_rotation("angled")
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		if rotation == 'axial':
			rotation = [0,0,0]
		elif rotation == 'sagittal':
			rotation = [0,90,-90]
		elif rotation == 'coronal':
			rotation = [-90,0,0]
		elif rotation == 'angled':
			rotation = [22.5,22.5,225]
		
		rotation = utils.sanitize_vector3(rotation)
		self.rotation = rotation
		client.sio.emit('SetCameraRotation', {self.id: rotation})

	def set_zoom(self,zoom):
		"""Set the camera zoom. 

		Parameters
		----------
		zoom : float	
			camera zoom parameter

		Examples
		--------
		>>> c1.set_zoom(1.0)
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		self.zoom = zoom
		client.sio.emit('SetCameraZoom', {self.id: zoom})

	def set_target_area(self, camera_target_area):
		"""Set the camera rotation to look towards a target area

		Note: some long/curved areas have mesh centers that aren't the 'logical' center. For these areas, calculate a center yourself and use set_camera_target.

		Parameters
		----------
		camera_target_area : string
			area ID or acronym, append "-lh" or "-rh" for one-sided meshes
		Examples
		--------
		>>> c1.set_target_area("grey-l") 
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		camera_target_area
		self.target = camera_target_area
		client.sio.emit('SetCameraTargetArea', {self.id: camera_target_area})

	def set_pan(self,pan_x, pan_y):
		"""Set camera pan coordinates

		Parameters
		----------
		pan_x : float
			x coordinate
		pan_y : float
			y coordinate
		
		Examples
		--------
		>>> c1.set_pan(3.0, 4.0)
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		self.pan = [pan_x, pan_y]
		client.sio.emit('SetCameraPan',{self.id: self.pan})

	def set_mode(self, mode):
		"""Set camera perspective mode

		Parameters
		----------
		mode : string
			"perspective" or "orthographic" (default)
		
		Examples
		--------
		>>> c1.set_mode('perspective')
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		self.mode = mode
		client.sio.emit('SetCameraMode', {self.id: mode})

	def set_background_color(self, background_color):
		"""Set camera background color

		Parameters
		----------
		background_color : hex color string

		Examples
		--------
		>>> c1.set_background_color('#000000') # black background
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		self.background_color = utils.sanitize_color(background_color)

		client.sio.emit('SetCameraColor', {self.id: self.background_color})

	def set_controllable(self):
		"""Sets camera to controllable
		
		Examples
		--------
		>>> c1.set_controllable()
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		self.controllable = True
		client.sio.emit('SetCameraControl', self.id)
		
	async def screenshot(self, size=[1024,768], filename = 'return'):
		"""Capture a screenshot

		Parameters
		----------
		size : list, optional
			Size of the screenshot, by default [1024,768]
		filename: string, optional
			Filename to save to, relative to local path
		"""
		global receive_totalBytes, receive_bytes, receive_camera
		self.image_received_event.clear()
		self.image_received = False
		receive_camera[self.id] = self

		if size[0] > 15000 or size[1] > 15000:
			raise Exception('(urchin.camera) Screenshots can''t exceed 15000x15000')
		
		client.sio.emit('RequestCameraImg', json.dumps({"name":self.id, "size":size}))

		while not self.image_received:
			await asyncio.sleep(0.1)
		# await self.image_received_event.wait()

		# image is here, reconstruct it
		img = Image.open(io.BytesIO(receive_bytes[self.id]))
		
		print(f'(Camera receive) {self.id} complete')
		del receive_totalBytes[self.id]
		del receive_bytes[self.id]
		del receive_camera[self.id]

		if not filename == 'return':
			img.save(filename)
		else:
			return img


def set_light_rotation(angles):
	"""Override the rotation of the main camera light

	Parameters
	----------
	angles : vector3
		Euler angles of light
	"""
	angles = utils.sanitize_vector3(angles)
	print(angles)
	print(isinstance(angles,list))
	client.sio.emit('SetLightRotation', angles)

def set_light_camera(camera_name = None):
	"""Change the camera that the main light is linked to (the light will rotate the camera)

	Parameters
	----------
	camera_name : string, optional
		Name of camera to attach light to, by default None
	"""
	if (camera_name is None):
		client.sio.emit('ResetLightLink')
	else:
		client.sio.emit('SetLightLink', camera_name)
	
main = Camera(main = True)


"""
urchin.

Python library for connecting to and sending data to a Universal Renderer for Neuroscience renderer.
"""
__author__ = 'Daniel Birman'
__version__ = "0.5.0"

# load the client
from . import client
from .renderer import *

# load sanitization
from . import utils

# load the scene controls
from . import camera

# load the object controls
from . import lines
from . import meshes
from . import particles
from . import probes
from . import text
from . import volumes
from . import texture

# load the colors
from . import colors

# load the atlases
from .atlas import *



"""Sanitizing inputs to send through API"""
import numpy as np
from enum import Enum

### ENUMS
class Side(Enum):
    LEFT = -1
    FULL = 0
    RIGHT = 1

### SANITIZING FUNCTIONS

def sanitize_vector3(vector):
  """Guarantee that a vector is a vector 3, or raise an exception

  Parameters
  ----------
  input : any
      arbitrary input parameter

  Returns
  -------
  list
      vector3 as a list [x,y,z]

  Raises
  ------
  Exception
      Failed to coerce input to a length 3 list
  """
    #if statement to check type and then convert to vector3
  if(len(vector)!=3):
    raise Exception(f'Expected a vector3 but received length {len(vector)}.')
  
  if isinstance(vector, list):
      return vector
  
  if isinstance(vector, tuple) or isinstance(vector, np.ndarray):
      return list(vector)
  else:
      raise Exception('Vector3 failed to be coerced properly')



def sanitize_color(color):
    if isinstance(color, str): #if color is string
        if(color[0] == '#'): #if color has # prefix
            return(color)
        else: #adding # prefix if not present
            color = '#' + color
            return(color)    
    
    if isinstance(color, list) or isinstance(color, tuple):#if list
        # check first if you're dealing with floats between 0 and 1
        if isinstance(color[0], float):
            # we have floats, check which range we have
            if all([(x <= 1 and x >= 0) for x in color]):
                # range 0->1, re-scale to 0->255 as ints
                color = [int(255*x) for x in color]
            elif all([(x <= 255 and x >= 0) for x in color]):
                # range 0->255, round to ints
                color = [int(np.rint(x)) for x in color]
            else:
                raise Exception('Color value ranges need to conform to [R,G,B] formatted either as 0->1 or 0->255')

        # convert list of ints to a hex string
        hex_color = '#' + ('{:02X}' * 3).format(color[0], color[1], color[2])
        return(hex_color)
    
    else:
        raise Exception('Failed to re-format color input as a hex code. Please enter either an RGB list or tuple, or a hex code')

def sanitize_float(value):
    if isinstance(value, float):
        return value
    else:
        try:
            return float(value)
        except:
            raise Exception("Value could not be coerced to a float.")

def sanitize_material(material):
    if isinstance(material, str):
        return(material)
    else:
        raise Exception("Material is not properly passed in as a string. Please pass in material as a string.")

def sanitize_list(input, length=0):
    """Guarantee that a list is of at least size length, or try to broadcast to that size

    Parameters
    ----------
    input : list
    length : int, optional
        length to broadcast to, by default 0

    Returns
    -------
    list
    """
    if length > 0 and not isinstance(input, list):
        input = [input] * length

    if not isinstance(input, list):
        raise Exception("List parameter needs to be a list.")

    return input
    
def sanitize_string(string):
    if isinstance(string, str):
        return(string)
    else:
        raise Exception("Input is not properly passed in as a string. Please pass in input as a string.")
    
def sanitize_side(acronym, sided):
    if sided == "full":
        return acronym
    elif sided == "left":
        return f'{acronym}-lh'
    elif sided == "right":
        return f'{acronym}-rh'
    else:
        raise Exception(f'Sided enum {sided} not properly defined, should be full/left/right')


from .. import client
from .. import utils
from pathlib import Path

import json

class Atlas:
    def __init__(self, atlas_name):
        # load the ontology structure file
        self.atlas_name = atlas_name
        self.loaded = False

        current_script_directory = Path(__file__).resolve().parent
        
        data_file_path = f'{current_script_directory}/data/{atlas_name}.structures.json'
        with open(data_file_path,'r') as f:
            temp = json.load(f)

        self.areas = []
        for structure_data in temp:
            self.areas.append(structure_data['acronym'])
            structure_data['color'] = utils.sanitize_color(structure_data['rgb_triplet'])
            setattr(self, structure_data['acronym'], Structure(structure_data))

    def load(self):
        """Load this atlas
        """
        if self.loaded:
            print("(Warning) Atlas was already loaded, the renderer can have issues if you try to load an atlas twice.")
        
        self.loaded = True
        client.sio.emit('LoadAtlas', self.atlas_name)

    def clear(self):
        """Clear all visible areas
        """
        client.sio.emit('Clear', 'area')

    def load_defaults(self):
        """Load the left and right areas
        """
        client.sio.emit('LoadDefaultAreas', "")

    def get_areas(self, area_list):
        """Get the area objects given a list of area acronyms

        Parameters
        ----------
        area_list : list of string
            List of acronyms to get objects for

        Returns
        -------
        list of areas

        Examples
        --------
        >>> area_list = urchin.get_areas(["root", "VISp"])
        """
        areas = []
        for name in area_list:
            try:
                area = getattr(self, name)
                areas.append(area)
            except:
                print(f'(Warning): Area {name} couldn''t be found in this atlas!')
        return areas

    def set_visibilities(self, area_list, area_visibility, side = utils.Side.FULL):
        """Set visibility of multiple areas at once

        Parameters
        ----------
        area_list : list of string
            List of acronyms
        area_visibility : list of bool
        sided : string, optional
            Brain area side to control "full"/"left"/"right", default = "full"

        Examples
        --------
        >>> urchin.ccf25.set_visibilities(urchin.ccf25.get_areas(["root", "VISp"]), [True, False])
        >>> urchin.ccf25.set_visibilities(urchin.ccf25.get_areas(["root", "VISp"]), True)
        """
        area_visibility = utils.sanitize_list(area_visibility, len(area_list))

        # output dictionary should match JSON schema AreaData:
        #{"acronym": ["a", "b", "c"], "side": [-1, 0, 1], "visible": [true, true, false]}
        data_dict = {}
        data_dict['acronym'] = []
        data_dict['side'] = []
        data_dict['visible'] = []
        for i, area in enumerate(area_list):
            data_dict['acronym'].append(area.acronym)
            data_dict['visible'].append(area_visibility[i])
            data_dict['side'].append(side.value)

        client.sio.emit('SetAreaVisibility', json.dumps(data_dict))

    def set_colors(self, area_list, area_colors, sided="full"):
        """Set color of multiple areas at once.

        Parameters
        ----------
        area_colors : list of colors (hex string or RGB triplet)

        Examples
        --------
        >>> urchin.ccf25.set_visibilities(urchin.ccf25.get_areas(["root", "VISp"]), ['#ff0000', '#00ff00'])
        >>> urchin.ccf25.set_visibilities(urchin.ccf25.get_areas(["root", "VISp"]), [255, 255, 255])
        """
        area_colors = utils.sanitize_list(area_colors, len(area_list))
        for i in range(0,len(area_colors)):
            area_colors[i] = utils.sanitize_color(area_colors[i])

        data_dict = {}
        for i, area in enumerate(area_list):
            area_name = utils.sanitize_side(area.acronym, sided)
            data_dict[area_name] = area_colors[i]

        client.sio.emit('SetAreaColors', data_dict)
        
    def set_colormap(self, colormap_name):
        """Set colormap used for mapping area *intensity* values to colors


        Options are
        - cool (default, teal 0 -> magenta 1)
        - grey (black 0 -> white 1)
        - grey-green (grey 0, light 1/255 -> dark 1)
        - grey-purple (grey 0, light 1/255 -> dark 1)
        - grey-red (grey 0, light 1/255 -> dark 1)
        - grey-rainbow (grey 0, rainbow colors from 1/255 -> 1)

        Parameters
        ----------
        colormap_name : string
            colormap name
        """
        client.sio.emit('SetAreaColormap', colormap_name)

    def set_color_intensity(self, area_list, area_intensities, sided="full"):
        """Set intensity values, colors will be set according to the active colormap

        Parameters
        ----------
        area_intensities : list of float
            0->1

        Examples
        --------
        >>> urchin.ccf25.set_color_intensity(urchin.ccf25.get_areas(["root", "VISp"]), [0, 1])
        """
        area_intensities = utils.sanitize_list(area_intensities, len(area_list))
        for i in range(0,len(area_intensities)):
            area_intensities[i] = utils.sanitize_float(area_intensities[i])

        data_dict = {}
        for i, area in enumerate(area_list):
            area_name = utils.sanitize_side(area.acronym, sided)
            data_dict[area_name] = area_intensities[i]

        client.sio.emit('SetAreaIntensity', data_dict)

    def set_alphas(self, area_list, area_alphas, sided="full"):
        """Set transparency of multiple areas at once. Requires a transparent material. 
        
        Parameters
        ----------
        area_alphas : list of float
            Alpha ranges from 0->1

        Examples
        --------
        >>> urchin.ccf25.set_alphas(urchin.ccf25.get_areas(["root", "VISp"]), [0.15, 0.15])
        >>> urchin.ccf25.set_alphas(urchin.ccf25.get_areas(["root", "VISp"]), [0.5])
        """
        area_alphas = utils.sanitize_list(area_alphas, len(area_list))
        for i in range(0,len(area_alphas)):
            area_alphas[i] = utils.sanitize_float(area_alphas[i])

        data_dict = {}
        for i, area in enumerate(area_list):
            area_name = utils.sanitize_side(area.acronym, sided)
            data_dict[area_name] = area_alphas[i]
        
        client.sio.emit('SetAreaAlpha', data_dict)

    def set_materials(self, area_list, area_materials, sided="full"):
        """Set material of multiple areas at once.

        Material options are
        - 'opaque-lit' or 'default'
        - 'opaque-unlit'
        - 'transparent-lit'
        - 'transparent-unlit'
        
        Parameters
        ----------
        area_colors : dict {string: string}
            Keys are area IDs or acronyms, Values are hex colors

        Examples
        --------
        >>> urchin.ccf25.set_materials(urchin.ccf25.get_areas(["root", "VISp"]), ['transparent-lit', 'opaque-lit'])
        >>> urchin.ccf25.set_materials(urchin.ccf25.get_areas(["root", "VISp"]), ['transparent-lit])
        """
        area_materials = utils.sanitize_list(area_materials, len(area_list))
        for i in range(0,len(area_materials)):
            area_materials[i] = utils.sanitize_string(area_materials[i])

        data_dict = {}
        for i, area in enumerate(area_list):
            area_name = utils.sanitize_side(area.acronym, sided)
            data_dict[area_name] = area_materials[i]

        client.sio.emit('SetAreaMaterial', data_dict)

class Structure:
    """Structure attributes can be accessed as

    >>> structure.name
    >>> structure.acronym
    >>> structure.id
    >>> structure.rgb_triplet
    >>> structure.path
    """
    def __init__(self, structure_data):
        for key, value in structure_data.items():
            setattr(self, key, value)

    def set_test(self):
        print(self)

    def set_visibility(self, visibility, sided = utils.Side.FULL):
        """Set area visibility

        Parameters
        ----------
        visibility : bool
        sided : str, optional
            "full" "left" or "right, by default "full"

        Examples
        --------
        >>> urchin.ccf25.root.set_visibility(True)
        >>> urchin.ccf25.root.set_visibility(True, urchin.utils.Side.LEFT)
        """
        data_dict = {
            'acronym':[self.acronym],
            'side':[sided.value],
            'visible':[visibility]
        }

        client.sio.emit('SetAreaVisibility', json.dumps(data_dict))

    def set_color(self, color, sided = "full"):
        """Set area color.

        Parameters
        ----------
        color : hex string

        Examples
        --------
        >>> urchin.ccf25.root.set_color('#ff0000')
        >>> urchin.ccf25.root.set_color([255, 0, 0], "left")
        """
        color = utils.sanitize_color(color)

        area_name = utils.sanitize_side(self.acronym, sided)

        client.sio.emit('SetAreaColors', {area_name:color})

    def set_intensity(self, intensity, sided = "full"):
        """Set color based on the intensity value through the active colormap.

        Parameters
        ----------
        intensity : float
            0->1

        Examples
        --------
        >>> urn.set_intensity(0.5)
        """
        intensity = utils.sanitize_float(intensity)
        area_name = utils.sanitize_side(self.acronym, sided)

        client.sio.emit('SetAreaIntensity', {area_name:intensity})

    def set_alpha(self, alpha, sided = "full"):
        """Set transparency. 

        Parameters
        ----------
        alpha : float

        Examples
        --------
        >>> urchin.ccf25.root.set_alpha(0.15)
        >>> urchin.ccf25.root.set_alpha(0.5, "left")
        """
        alpha = utils.sanitize_float(alpha)

        area_name = utils.sanitize_side(self.acronym, sided)

        client.sio.emit('SetAreaAlpha', {area_name:alpha})

    def set_material(self, material, sided = "full"):
        """Set material.

        Material options are
        - 'opaque-lit' or 'default'
        - 'opaque-unlit'
        - 'transparent-lit'
        - 'transparent-unlit'

        Parameters
        ----------
        material: string

        Examples
        ----------
        >>> urchin.ccf25.root.set_material('transparent-lit')
        """
        material = utils.sanitize_string(material)

        area_name = utils.sanitize_side(self.acronym, sided)

        client.sio.emit('SetAreaMaterial', {area_name:material})

    # def set_data(area_data):
    #     """Set the data array for each CCF area model

    #     Data arrays work the same as the set_area_intensity() function but are controlled by the area_index value, which can be set in the renderer or through the API.

    #     Parameters
    #     ----------
    #     area_data : dict {string: float list}
    #         keys area IDs or acronyms, values are a list of floats
    #     """
    #     client.sio.emit('SetAreaData', area_data)

    # def set_data_index(area_index):
    #     """Set the data index for the CCF area models

    #     Parameters
    #     ----------
    #     area_index : int
    #         data index
    #     """
    #     client.sio.emit('SetAreaIndex', area_index)


"""Particles"""

from . import client
import warnings
from . import utils

## Particle system
counter = 0
class Particle:
	"""Particles should not be directly instantiated, use urchin.particles.create(n) and urchin.clear_particles()
	"""
	def __init__(self):
		pass

	def set_position(self, position):
		"""Set the position of a particle in ap/ml/dv coordinates relative to the origin (0,0,0)
		
		Parameters
		---------- 
		position : list of three floats
			(ap, ml, dv) coordinates in um

		Examples
		--------
		>>>p1.set_position([5200,5700,330]) # set a particle to Bregma, in CCF space
		"""
		if self.in_unity == False:
			raise Exception("Particle does not exist in Unity, call create method first.")
		
		self.position = utils.sanitize_vector3(position)
		client.sio.emit('SetParticlePos', {self.id: [self.position[0]/1000, self.position[1]/1000, self.position[2]/1000]})

	def set_size(self, size):
		"""Set the size of a particle
		
		Parameters
		---------- 
		size : float

		Examples
		--------
		>>>p1.set_size(0.02) # 20 um 
		"""
		if self.in_unity == False:
			raise Exception("Particle does not exist in Unity, call create method first.")
		
		size = utils.sanitize_float(size)
		self.size = size
		client.sio.emit('SetParticleSize', {self.id: size})
	
	def set_color(self, color):
		"""Set the color of a particle
		
		Parameters
		---------- 
		color : string hex color

		Examples
		--------
		>>>p1.set_color('#FFFFFF')
		"""
		if self.in_unity == False:
			raise Exception("Particle does not exist in Unity, call create method first.")
		
		color = utils.sanitize_color(color)
		self.color = color
		client.sio.emit('SetParticleColor', {self.id: color})

	# todo: re-implement shape and material when we move to ParticleGroup instead of single Particle objects

	# def set_shape(self, shape):
	# 	"""Set the shape of a particle
	# 	Options are
	#  	 - 'sphere' (default)
	#  	 - 'cube' better performance when rendering tens of thousands of neurons

	# 	Parameters
	# 	---------- 
	# 	shape : string
	# 		new shape of the neuron

	# 	Examples
	# 	--------
	# 	>>>p1.set_shape('sphere')
	# 	"""
	# 	if self.in_unity == False:
	# 		raise Exception("Particle does not exist in Unity, call create method first.")
		
	# 	self.shape = shape
	# 	client.sio.emit('SetNeuronShape', {self.id: shape})

	# def set_material(self, material):
	# 	"""Set the material of neuron renderer
	# 	Options are
	#  	- 'lit-transparent' (default)
	#  	- 'lit'
	#  	- 'unlit'

	# 	Parameters
	# 	---------- 
	# 	material : string
	# 		new material of the neuron

	# 	Examples
	# 	--------
	# 	>>>p1.set_material('lit-transparent')
	# 	"""
	# 	if self.in_unity == False:
	# 		raise Exception("Particle does not exist in Unity, call create method first.")
		
	# 	self.material = material
	# 	client.sio.emit('SetNeuronMaterial', {self.id: material})

def create(num_particles):
	"""Create particles

	Note: particles must be created before setting other values

	Parameters
	----------
	num_particles : int

	Examples
	--------
	>>> neurons = urchin.particles.create(3)
	"""
	global counter
	neurons = []
	for i in range(num_particles):
		counter += 1
		particle = Particle()
		particle.id = f'n{str(counter)}'
		particle.in_unity = True
		neurons.append(particle)

	client.sio.emit('CreateParticles', [x.id for x in neurons])
	return neurons

def clear():
	"""Clear all particles

	Note that there is no delete method for individual particles, they must all be cleared at once.
	"""
	client.sio.emit('Clear', 'particle')

def set_positions(particles_list, positions_list):
	"""Set the position of particles in ap/ml/dv coordinates relative to the origin

	Parameters
	----------
	particles_list : list of Particle
	positions_list : list of list of three floats
		list of positions of neurons (ap, ml, dv) in um

	Examples
	--------
	>>> urchin.particles.set_positions([p1,p2,p3], [[1000,1000,1000],...,...])
	"""
	particles_list = utils.sanitize_list(particles_list)
	positions_list = utils.sanitize_list(positions_list)

	neuron_positions = {}
	for i in range(len(particles_list)):
		neuron = particles_list[i]
		if neuron.in_unity:
			pos = utils.sanitize_vector3(positions_list[i])
			neuron_positions[neuron.id] = [pos[0]/1000, pos[1]/1000, pos[2]/1000]
		else:
			warnings.warn(f"Particle with id {neuron.id} does not exist in Unity, call create method first.")
	client.sio.emit('SetParticlePos', neuron_positions)

def set_sizes(particles_list, sizes_list):
	"""Set particles sizes

	Parameters
	----------
	particles_list : list of Particle
	sizes_list : list of float

	Examples
	--------
	>>> urchin.particles.set_sizes([p1,n2,n3], [0.01,0.02,0.03])
	"""
	particles_list = utils.sanitize_list(particles_list)
	sizes_list = utils.sanitize_list(sizes_list)

	neurons_sizes = {}
	for i in range(len(particles_list)):
		neuron = particles_list[i]
		if neuron.in_unity:
			neurons_sizes[neuron.id] = utils.sanitize_float(sizes_list[i])
		else:
			warnings.warn(f"Particle with id {neuron.id} does not exist in Unity, call create method first.")
	client.sio.emit('SetParticleSize', neurons_sizes)

def set_colors(particles_list, colors_list):
	"""Set neuron colors

	Parameters
	----------
	particles_list : list of Particle
	colors_list : list of string hex colors
		list of colors of neurons

	Examples
	--------
	>>> urchin.particles.set_colors([p1,n2,n3], ['#FFFFFF','#000000','#FF0000'])
	"""
	particles_list = utils.sanitize_list(particles_list)
	colors_list = utils.sanitize_list(colors_list)

	neurons_colors = {}
	for i in range(len(particles_list)):
		neuron = particles_list[i]
		if neuron.in_unity:
			neurons_colors[neuron.id] = utils.sanitize_color(colors_list[i])
		else:
			warnings.warn(f"Particle with id {neuron.id} does not exist in Unity, call create method first.")
	client.sio.emit('SetParticleColor', neurons_colors)

def set_material(material_name):
	"""Change the material used to render neurons

 	Options are
 	- 'gaussian' (default)
 	- 'circle'
 	- 'circle-lit'
 	- 'square'
 	- 'square-lit'
 	- 'diamond'
 	- 'diamond-lit'

	Parameters
	----------
	material_name: string

	Examples
	--------
	>>> urchin.neurons.set_material('circle')
	"""
	material_name = utils.sanitize_string(material_name)

	client.sio.emit('SetParticleMaterial', material_name)




"""Primitive meshes"""

from . import client
import warnings
from . import utils

  ## Primitive Mesh Renderer
counter = 0
class Mesh:
  #Run everytime an object is created, sets the fields to defaults if one is not given, and sends the info to Unity
  #id = int index counter
  def __init__(self,position= [0.0,0.0,0.0], scale= [1,1,1], color= '#FFFFFF', material = 'default'):
    self.create()

    
    position = utils.sanitize_vector3(position)
    self.position = position
    client.sio.emit('SetPosition', {self.id: position})
      
    
    scale = utils.sanitize_vector3(scale)
    self.scale = scale
    client.sio.emit('SetScale', {self.id: scale})
      
   
    color = utils.sanitize_color(color)
    self.color = color
    client.sio.emit('SetColor',{self.id: color})

    material = utils.sanitize_material(material)
    self.material = material
    

    
  def create(self):
    """Creates primitive mesh

    Parameters
    ----------
    none
      
	  Examples
	  --------
	  >>> cube_obj = urchin.primitives.Primitive()
    """
    global counter
    counter +=1
    self.id = 'prim' + str(counter)
    client.sio.emit('CreateMesh', [self.id])
    self.in_unity = True
  


  def delete(self):
    """Deletes meshes

    Parameters
    ----------
    references object being deleted
      
	  Examples
	  --------
	  >>> cube_obj.delete() 
    """
    client.sio.emit('DeleteMesh', [self.id])
    self.in_unity = False
  
  def set_position(self, position):
    """Set the position of mesh renderer

    Parameters
    ----------
    position : list of three floats
      vertex positions of the mesh
        
    Examples
    --------
    >>> cube_obj.set_position([2,2,2])
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    
    position = utils.sanitize_vector3(position)
    self.position = position
    client.sio.emit('SetPosition', {self.id: position})
  
  def set_scale(self, scale):
    """Set the scale of mesh renderer

    Parameters
    ----------
    scale : list of three floats
      new scale of mesh
        
    Examples
    --------
    >>> cube_obj.set_scale([3,2,3])
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    
    scale = utils.sanitize_vector3(scale)
    self.scale = scale
    client.sio.emit('SetScale', {self.id: scale})
  
  def set_color(self, color):
    """Set the color of mesh renderer

    Parameters
    ----------
    mesh_color : string hex color
      new hex color of mesh
        
    Examples
    --------
    >>> cube_obj.set_color("#000000")
    
    """ 
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    
    color = utils.sanitize_color(color)
    self.color = color
    client.sio.emit('SetColor',{self.id: color})

  def set_material(self, material):
    """Set the material of mesh renderer

    Parameters
    ----------
    mesh_material : string
      name of new material
        
    Examples
    --------
    >>> cube_obj.set_material('opaque-lit')
    
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    
    material = utils.sanitize_material(material)
    self.material = material
    client.sio.emit('SetMaterial',{self.id: material})




#actually initializes each object(s), doesn't use any parameters other than how many to initialize (uses all defaults)
def create(num_objects):
  """Creates primitive meshes

  Parameters
  ----------
  num_objects : int
    number of primitive objects to be created      
	Examples
	--------
	>>> cubes = urchin.primitives.create(2)
  """
  prim_names = []
  for i in range(num_objects):
    prim_names.append(Mesh())
  return(prim_names)

def delete(meshes_list):
  """Deletes meshes

  Parameters
  ----------
  meshes_list : list of mesh objects
	  list of meshes being deleted
      
	Examples
	--------
	>>> cubes.delete()
  """
  meshes_list = utils.sanitize_list(meshes_list)

  mesh_ids = [x.id for x in meshes_list]
  client.sio.emit('DeleteMesh', mesh_ids)

def set_positions(meshes_list, positions_list):
  """Set the positions of mesh renderers

  Parameters
  ----------
  meshes_list : list of mesh objects
	  list of meshes being set
  positions_list : list of list of three floats
    vertex positions of each mesh
      
	Examples
	--------
	>>> urchin.primitives.set_positions(cubes,[[3,3,3],[2,2,2]])
  """
  meshes_list = utils.sanitize_list(meshes_list)
  positions_list = utils.sanitize_list(positions_list)

  mesh_pos = {}
  for i in range(len(meshes_list)):
    mesh = meshes_list[i]
    if mesh.in_unity:
      mesh_pos[mesh.id] = utils.sanitize_vector3(positions_list[i])
    else:
      warnings.warn(f"Object with id {mesh.id} does not exist. Please create object {mesh.id}.")

  client.sio.emit('SetPosition', mesh_pos)

def set_scales(meshes_list, scales_list):
  """Set scale of mesh renderers

  Parameters
  ----------
  meshes_list : list of mesh objects
	  list of meshes being scaled
  scales_list : list of list of three floats
    new scales of each mesh
      
	Examples
	--------
	>>> urchin.primitives.set_scales(cubes,[[3,3,3],[2,2,2]])
  """
  meshes_list = utils.sanitize_list(meshes_list)
  scales_list = utils.sanitize_list(scales_list)

  mesh_scale = {}
  for i in range(len(meshes_list)):
    mesh = meshes_list[i]
    if mesh.in_unity:
      mesh_scale[mesh.id] = utils.sanitize_vector3(scales_list[i])
    else:
      warnings.warn(f"Object with id {mesh.id} does not exist. Please create object {mesh.id}.")

  client.sio.emit('SetScale', mesh_scale)

def set_colors(meshes_list, colors_list):
  """Sets colors of mesh renderers

  Parameters
  ----------
  meshes_list : list of mesh objects
	  list of meshes undergoing color change
  colors_list : list of string hex colors
      new hex colors for each mesh
      
	Examples
	--------
	>>> urchin.primitives.set_colors(cubes,["#000000","#000000"])
	
  """
  meshes_list = utils.sanitize_list(meshes_list)
  colors_list = utils.sanitize_list(colors_list)

  mesh_colors = {}
  for i in range(len(meshes_list)):
    mesh = meshes_list[i]
    if mesh.in_unity:
      mesh_colors[mesh.id] = utils.sanitize_color(colors_list[i])
    else:
      warnings.warn(f"Object with id {mesh.id} does not exist. Please create object {mesh.id}.")

  client.sio.emit('SetColor', mesh_colors)

def set_materials(meshes_list, materials_list):
  """Sets materials of mesh renderers

  Parameters
  ----------
  meshes_list : list of mesh objects
	  list of meshes undergoing material change
  materials_list : list of strings
      name of new materials
      
	Examples
	--------
	>>> urchin.primitives.set_material(cubes,['opaque-lit','opaque-lit'])
	
  """
  meshes_list = utils.sanitize_list(meshes_list)
  materials_list = utils.sanitize_list(materials_list)

  mesh_materials = {}
  for i in range(len(meshes_list)):
    mesh = meshes_list[i]
    if mesh.in_unity:
      mesh_materials[mesh.id] = utils.sanitize_material(materials_list[i])
    else:
      warnings.warn(f"Object with id {mesh.id} does not exist. Please create object {mesh.id}.")  
      
  client.sio.emit('SetMaterial', mesh_materials) 


"""Calcium"""
from . import client
import warnings
from . import utils
from PIL import Image
from typing import List
import io

receive_fname = ''
receive_count = 0
receive_data = []

CHUNK_SIZE = 100000000

counter = 0
class Texture:
    def __init__(self, position=None, offset=None, texture_file=None):
        self.create()

        if position: self.set_position(position)
        if offset: self.set_offset(offset)
        if texture_file: self.set_image(texture_file)

    def create(self):
        """Creates Textures

        Parameters
        ---------- 
        none

        Examples
        >>> tex = urchin.texture.Texture()
        """
        global counter
        counter += 1
        self.id = 'tex' + str(counter)
        client.sio.emit('CreateFOV',[self.id])
        self.in_unity = True

    def delete(self):
        """Deletes Textures

        Parameters
        ---------- 
        references object being deleted

        Examples
        >>> tex.delete()
        """
        client.sio.emit('DeleteFOV',[self.id])
        self.in_unity = False

    def set_position(self,positions):
        """Set the position of fov in ap/ml/dv coordinates relative to the CCF (0,0,0) point

        Parameters
        ---------- 
        position : list of list of three floats
        	vertex positions of the four corners of the texture relative to the CCF origin

        Examples
        --------
        >>> tex.set_position([[9.2611688 , 7.94252654, 1.3904818 ],
                [8.75443871, 7.94252654, 1.08008647],
                [9.24259926, 8.52829154, 1.43690564],
                [8.73586918, 8.52829154, 1.12651031]])
        """
        if self.in_unity == False:
            raise Exception("Texture does not exist in Unity, call create method first.")

        for i, pos in enumerate(positions):
            positions[i] = utils.sanitize_vector3(pos)

        self.position = utils.sanitize_list(positions)
        client.sio.emit('SetFOVPos',{self.id: positions})
    
    def set_image(self, array):
        """Set the image data for texture

        Parameters
        ----------
        array : numpy array
            luminance data for texture

        Examples
        --------
        >>> tex.set_texture(array)
        """
        if self.in_unity == False:
            raise Exception("Texture does not exist in Unity, call create method first.")

        # texture_file = utils.sanitize_string(texture_file)
        # self.texture_file=texture_file

        # img = Image.open(texture_file)
        # Convert img to bytes
        img_bytes = array.tobytes()
        # img_bytes.seek(0)
        # Split bytes into chunks
        chunks = [img_bytes]
            
        client.sio.emit('SetFOVTextureDataMetaInit', [self.id, len(chunks), array.shape[0], array.shape[1], 'array'])

        # Send img by chunk
        # [TODO: Replace with a data structure]
        for i,chunk in enumerate(chunks):
            immediate_apply = True if i==len(chunks)-1 else False
            client.sio.emit('SetFOVTextureDataMeta', [self.id,i,immediate_apply])
            client.sio.emit('SetFOVTextureData',chunk)

    def set_offset(self, offset):
        """Set the vertical offset for this texture

        Parameters
        ----------
        offset : float
            Vertical offset in mm
        """
        client.sio.emit('SetFOVOffset', {self.id: offset})


def create(N):
    """Create Texture objects

    Note: textures must be created before setting other values

    Parameters
    ----------
    N : int
        number of new Texture objects

    Examples
    --------
    >>> textures_list = urchin.textures.create(3)
    """
    textures = []
    for _ in range(N):
        textures.append(Texture())
    return textures

def delete(textures_list):
    """Delete Texture objects

    Parameters
    ----------
    textures_list : list of Texture

    Examples
    --------
    >>> urchin.fovs.delete(textures_list)
    """
    textures_list = utils.sanitize_list(textures_list)
    for tex in textures_list:
        if tex.in_unity:
            tex.delete()
        else:
            warnings.warn(f"fov with id {tex.id} does not exist in Unity, call create method first.")

    fovs_ids = [x.id for x in textures_list]
    client.sio.emit('DeleteFOVs', fovs_ids)

def set_positions(textures_list, positions_list):
    """Set the positions of textures in ap/ml/dv coordinates relative to the CCF (0,0,0) point

    Note: this plural function has no efficiency advantage of Texture.set_position, for readable
    code we recommend using the former.

    Parameters
    ----------
    textures_list : list of fov objects
        list of fovs being moved
    positions : list of list of list of three floats
        list of four vertex corners for each texture

    Examples
    --------
    >>> urchin.fovs.set_positions([[...],[...],[...]]) # see set_position for position example
    """
    textures_list = utils.sanitize_list(textures_list)
    positions_list = utils.sanitize_list(positions_list)

    for fov,position in zip(textures_list, positions_list):
        if fov.in_unity:
            fov.set_position(position)
        else:
            warnings.warn(f"fov with id {fov.id} does not exist in Unity, call create method first.")

def set_images(textures_list, images_list):
    """Set the position of fov in ml/ap/dv coordinates relative to the CCF (0,0,0) point

    Parameters
    ----------
    fovs_list : list of fov objects
        list of fovs being moved
    positions : list of list of three floats
        list of positions of fovs

    Examples
    --------
    >>> img0 = fov.
    >>> urchin.fovs.set_textures([f1,f2,f3], ['fov1.png','fov'])
    """
    textures_list = utils.sanitize_list(textures_list)
    images_list = utils.sanitize_list(images_list, len(textures_list))

    for tex,img in zip(textures_list,images_list):
        if tex.in_unity:
            tex.set_texture(img)
        else:
            warnings.warn(f"Texture with id {tex.id} does not exist in Unity, call create method first.")


"""Lines"""

from . import client
import warnings
from . import utils


counter = 0

class Line:
  def __init__(self, position= [[0.0,0.0,0.0]], color= '#FFFFFF'):
    self.create()

    self.set_position(position)
    self.set_color(color)

  def create(self):
    """Creates lines
    
    Parameters
    ---------- 
    none

    Examples
    >>>l1 = urchin.lines.Line()
    """
    global counter
    counter += 1
    self.id = 'l' + str(counter)
    client.sio.emit('CreateLine', [self.id])
    self.in_unity = True

  def delete(self):
    """Deletes lines
    
    Parameters
    ---------- 
    references object being deleted

    Examples
    >>>l1.delete()
    """
    client.sio.emit('DeleteLine', [self.id])
    self.in_unity = False

  def set_position(self, position):
    """Set the position of line renderer
    
    Parameters
    ---------- 
    position : list of vector3 [[ap,ml,dv],[ap,ml,dv]]
        vertex positions of the line in the ReferenceAtlas space (um)

    Examples
    --------
    >>>l1.set_position([[0, 0, 0],[13200,11400,8000]])
    """
    if self.in_unity == False:
      raise Exception("Line does not exist in Unity, call create method first.")

    for i, vec3 in enumerate(position):
      position[i] = utils.sanitize_vector3(vec3)
    self.position = position

    client.sio.emit('SetLinePosition', {self.id: self.position})

  def set_color(self, color):
    """Set the color of line renderer
    
    Parameters
    ---------- 
    color : string hex color
        new color of the line

    Examples
    --------
    >>>l1.set_color('#000000')
    """
    if self.in_unity == False:
      raise Exception("Line does not exist in Unity, call create method first.")

    color = utils.sanitize_color(color)
    self.color = color
    client.sio.emit('SetLineColor',{self.id: color})

def create (n):
  """Create Line objects

  Parameters
  ----------
  n : int
      Number of objects to create
  """
  lines_list = []
  
  for i in range(n):
    line = Line()
    lines_list.append(line)

  return lines_list

def delete (lines_list):
  """Deletes lines
  
  Parameters
  ---------- 
  lines_list : list of Line objects
      list of lines to be deleted

  Examples
  --------
  >>> lines.delete()
  """
  lines_list = utils.sanitize_list(lines_list)
  lines_ids = [x.id for x in lines_list]
  client.sio.emit("DeleteLine", lines_ids)
